from dataclasses import dataclass
from typing import List

import wpilib


DRIVER_CONTROLLER_ID = 0
OPERATOR_CONTROLLER_ID = 1


class Joysticks:
    joysticks: list[wpilib.Joystick] = [
        wpilib.Joystick(DRIVER_CONTROLLER_ID),
        wpilib.Joystick(OPERATOR_CONTROLLER_ID)
    ]


@dataclass
class JoystickAxis:
    controller_id: int
    axis_id: int

    @property
    def value(self) -> float:
        return Joysticks.joysticks[self.controller_id].getRawAxis(self.axis_id)



from unum import Unum, units

m = units.m
ft = Unum.unit("ft", 0.3048 * m, "foot")
inch = Unum.unit("in", ft / 12, "inch")
mile = units.mile

s = units.s
ms = units.ms
minute = units.min
hour = units.h

rad = units.rad
deg = units.deg
rev = Unum.unit("rev", 360 * deg, "revolution")




import pickle
import socket

from robotpy_toolkit_7407.subsystem import Subsystem
from robotpy_toolkit_7407.network.objects import RobotStatusPacket, NetworkObject
from robotpy_toolkit_7407.utils import logger


class Network:
    """
    ### SCHEMA:

    # Robot status:
    - UDP status packet is sent by the robot to python on the DS laptop (port 5800)
       - Pickled string of the RobotStatusPacket struct
    - Python backend on DS is responsible for unpickling, validation, and converting to JSON for the frontend

    # DS to Robot communication
    - Python backend responsible for converting the actions into packets
    - TCP packets on port 5801
    - Pickled string
    - Each distinct command will have a defined struct
    - Response codes??
    """

    subsystems: list[Subsystem]

    udp_ip: str = "127.0.0.1"
    udp_port: int = 5800
    udp_socket: socket.socket
    udp_max_packet_size: int = 32768

    local_websocket_ip: str = "localhost"
    local_websocket_port: int = 8765

    @classmethod
    def robot_init(cls, subsystems: list[Subsystem]):
        logger.info("initializing network on robot", "[network]")

        cls.subsystems = subsystems

        logger.info("initializing udp", "[network]")
        cls.udp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        logger.info("initialization complete", "[network]")

    @classmethod
    def ds_init(cls):
        logger.info("initializing network on driver station", "[network]")

        logger.info("initializing udp", "[network]")
        cls.udp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        cls.udp_socket.bind((cls.udp_ip, cls.udp_port))

        logger.info("initialization complete", "[network]")

    @classmethod
    def robot_send_status(cls):
        cls._udp_send_network_obj(cls._robot_get_status())

    @classmethod
    def ds_get_status(cls) -> RobotStatusPacket:
        while True:
            status = cls._udp_receive_network_obj()
            if isinstance(status, RobotStatusPacket):
                return status

    @classmethod
    def _robot_get_status(cls) -> RobotStatusPacket:
        subsystem_status = []

        for s in cls.subsystems:
            subsystem_status.append(s.get_network_object())

        return RobotStatusPacket(
            subsystems=subsystem_status
        )

    @classmethod
    def _udp_send_network_obj(cls, obj: NetworkObject):
        data = pickle.dumps(obj)
        if len(data) > cls.udp_max_packet_size:
            logger.warn(f"Network object {obj} exceeded maximum packet size and was not sent")
            return
        cls.udp_socket.sendto(data, (cls.udp_ip, cls.udp_port))

    @classmethod
    def _udp_receive_network_obj(cls) -> NetworkObject:
        data, _ = cls.udp_socket.recvfrom(cls.udp_max_packet_size)
        obj: NetworkObject = pickle.loads(data)
        return obj



import math

from unum import Unum

from robotpy_toolkit_7407.oi.joysticks import JoystickAxis
from robotpy_toolkit_7407.subsystem import Subsystem
from robotpy_toolkit_7407.utils import logger
from robotpy_toolkit_7407.utils.math import rotate_vector, bounded_angle_diff
from robotpy_toolkit_7407.utils.units import s, m, deg, rad, hour, mile, rev


class SwerveNode:
    _motor_reversed: bool
    _motor_sensor_offset: Unum = rad

    def init(self):
        self._motor_reversed = False
        self._motor_sensor_offset = 0 * rad

    def set(self, vel: Unum, angle_radians: Unum):
        self._set_angle(angle_radians, self.get_current_motor_angle() + self._motor_sensor_offset)
        self.set_motor_velocity(vel if not self._motor_reversed else -vel)

    # OVERRIDDEN FUNCTIONS
    def set_motor_angle(self, pos: Unum): ...
    def get_current_motor_angle(self) -> Unum: ...
    def set_motor_velocity(self, vel: Unum): ...
    def get_motor_velocity(self) -> Unum: ...

    # 0 degrees is facing right
    def _set_angle(self, target_angle: Unum, initial_angle: Unum):
        target_sensor_angle, flipped, flip_sensor_offset = SwerveNode._resolve_angles(target_angle, initial_angle)

        target_sensor_angle -= self._motor_sensor_offset

        if flipped:
            self._motor_reversed = not self._motor_reversed
            self._motor_sensor_offset += flip_sensor_offset

        self.set_motor_angle(target_sensor_angle)

    @staticmethod
    def _resolve_angles(target_angle: Unum, initial_angle: Unum) -> (float, bool, float):
        """
        :param target_angle: Target node angle
        :param initial_angle: Initial node sensor angle
        :return: (target_sensor_angle, flipped, flip_sensor_offset)
        """

        target_rad = target_angle.asNumber(rad)
        initial_rad = initial_angle.asNumber(rad)

        # Actual angle difference in radians
        diff = bounded_angle_diff(initial_rad, target_rad)

        # Should we flip
        if abs(diff) > 0.65 * math.pi:
            flip_sensor_offset = math.pi if diff > 0 else -math.pi
            diff -= flip_sensor_offset
            return (diff + initial_rad) * rad, True, flip_sensor_offset * rad

        return (diff + initial_rad) * rad, False, 0 * rad


class SwerveOdometry:
    def init(self): ...
    def get_robot_heading(self) -> Unum: ...
    def reset_angle(self): ...


class SwerveDrivetrain(Subsystem):
    n_00: SwerveNode  # Top Left
    n_01: SwerveNode  # Bottom Left
    n_10: SwerveNode  # Top Right
    n_11: SwerveNode  # Bottom Right
    odometry: SwerveOdometry
    axis_dx: JoystickAxis
    axis_dy: JoystickAxis
    axis_rotation: JoystickAxis
    track_width: Unum = 1 * m
    max_vel: Unum = 20 * mile/hour
    max_angular_vel: Unum = 4 * rev/s
    deadzone_velocity: Unum = 0.05 * m/s
    deadzone_angular_velocity: Unum = 5 * deg/s

    def init(self):
        logger.info("initializing swerve drivetrain", "[swerve_drivetrain]")
        self.n_00.init()
        self.n_01.init()
        self.n_10.init()
        self.n_11.init()
        self.odometry.init()
        logger.info("initialization complete", "[swerve_drivetrain]")

    def set(self, vel: (Unum, Unum), angular_vel: Unum):
        vel = rotate_vector(vel[0], vel[1], -self.odometry.get_robot_heading())

        # logger.info(f"ROBOT AT {self.odometry.get_robot_angle_degrees()} degrees offset")
        # logger.info(f"({vel_tw_per_second[0]} tw/sec, {vel_tw_per_second[1]} tw/sec, {angular_vel} rad/sec)")

        if abs(vel[0]) < self.deadzone_velocity and abs(vel[1]) < self.deadzone_velocity and \
                abs(angular_vel) < self.deadzone_angular_velocity:
            self.n_00.set_motor_velocity(0 * m/s)
            self.n_01.set_motor_velocity(0 * m/s)
            self.n_10.set_motor_velocity(0 * m/s)
            self.n_11.set_motor_velocity(0 * m/s)
            return

        self.n_00.set(*self._calculate_swerve_node(
            -.5 * self.track_width, -.5 * self.track_width,
            vel[0], vel[1], angular_vel
        ))
        self.n_01.set(*self._calculate_swerve_node(
            -.5 * self.track_width, .5 * self.track_width,
            vel[0], vel[1], angular_vel
        ))
        self.n_10.set(*self._calculate_swerve_node(
            .5 * self.track_width, -.5 * self.track_width,
            vel[0], vel[1], angular_vel
        ))
        self.n_11.set(*self._calculate_swerve_node(
            .5 * self.track_width, .5 * self.track_width,
            vel[0], vel[1], angular_vel
        ))

    def stop(self):
        self.n_00.set(0 * m/s, 0 * rad/s)
        self.n_01.set(0 * m/s, 0 * rad/s)
        self.n_10.set(0 * m/s, 0 * rad/s)
        self.n_11.set(0 * m/s, 0 * rad/s)

    @staticmethod
    def _calculate_swerve_node(node_x: Unum, node_y: Unum, dx: Unum, dy: Unum, d_theta: Unum) -> (Unum, Unum):
        tangent_x, tangent_y = -node_y, node_x
        tangent_m = math.sqrt(tangent_x.asNumber(m)**2 + tangent_y.asNumber(m)**2) * m
        tangent_x /= tangent_m / m
        tangent_y /= tangent_m / m

        r = math.sqrt(2) / 2
        sx = dx + r * d_theta * tangent_x
        sy = dy + r * d_theta * tangent_y

        sx_u = sx.asNumber(m/s)
        sy_u = sy.asNumber(m/s)

        theta = math.atan2(sy_u, sx_u) * rad
        magnitude = math.sqrt(sx_u ** 2 + sy_u ** 2) * m/s
        return magnitude, theta



from unum import Unum

from robotpy_toolkit_7407.motor import PIDMotor
from robotpy_toolkit_7407.oi.joysticks import JoystickAxis
from robotpy_toolkit_7407.subsystem import Subsystem
from robotpy_toolkit_7407.utils import logger
from robotpy_toolkit_7407.utils.units import rad, m


class DifferentialDrivetrain(Subsystem):
    m_left: PIDMotor = None
    m_right: PIDMotor = None
    axis_x: JoystickAxis = None
    axis_y: JoystickAxis = None
    gear_ratio: Unum = rad/m

    def init(self):
        logger.info("initializing differential drivetrain", "[differential_drivetrain]")
        self.m_left.init()
        self.m_right.init()
        logger.info("initialization complete", "[differential_drivetrain]")

    def set_motor_percent_output(self, left: float, right: float):
        self.m_left.set_raw_output(left)
        self.m_right.set_raw_output(right)

    def set_motor_velocity(self, left_vel: Unum, right_vel: Unum):
        self.m_left.set_target_velocity(left_vel * self.gear_ratio)
        self.m_right.set_target_velocity(right_vel * self.gear_ratio)



import math

from robotpy_toolkit_7407.command import SubsystemCommand
from robotpy_toolkit_7407.subsystem_templates.drivetrain.swerve_drivetrain import SwerveDrivetrain
from robotpy_toolkit_7407.utils.units import m, s


class DriveSwerve(SubsystemCommand[SwerveDrivetrain]):
    def initialize(self) -> None:
        pass

    def execute(self) -> None:
        dx, dy, d_theta = self.subsystem.axis_dx.value, self.subsystem.axis_dy.value, self.subsystem.axis_rotation.value

        # if abs(dx) < 0.1:
        #     dx = 0
        # if abs(dy) < 0.1:
        #     dy = 0
        # if abs(d_theta) < 0.1:
        #     d_theta = 0

        # TODO normalize this to circle somehow
        dx *= self.subsystem.max_vel.asUnit(m/s)
        dy *= -self.subsystem.max_vel.asUnit(m/s)

        self.subsystem.set((dx, dy), d_theta * self.subsystem.max_angular_vel)

    def end(self, interrupted: bool) -> None:
        self.subsystem.stop()

    def isFinished(self) -> bool:
        return False

    def runsWhenDisabled(self) -> bool:
        return False
